import numpy
import scipy
import numpy as np
class FIRFilters:
    '''
    This file defines the main physical constants of the system
    '''

    # Speed of sound c=343 m/s
    c = 343.

    # distance to the far field
    ffdist = 10.

    # cut-off frequency of standard high-pass filter
    fc_hp = 300.

    # tolerance for computations
    eps = 1e-10

    def to_16b(self, signal):
        '''
        converts float 32 bit signal (-1 to 1) to a signed 16 bits representation
        No clipping in performed, you are responsible to ensure signal is within
        the correct interval.
        '''
        return ((2 ** 15 - 1) * signal).astype(np.int16)

    def clip(self, signal, high, low):
        '''
        Clip a signal from above at high and from below at low.
        '''
        s = signal.copy()

        s[np.where(s > high)] = high
        s[np.where(s < low)] = low

        return s

    def normalize(self, signal, bits=None):
        '''
        normalize to be in a given range. The default is to normalize the maximum
        amplitude to be one. An optional argument allows to normalize the signal
        to be within the range of a given signed integer representation of bits.
        '''

        s = signal.copy()

        s /= np.abs(s).max()

        # if one wants to scale for bits allocated
        if bits is not None:
            s *= 2 ** (bits - 1)
            s = self.clip(signal, 2 ** (bits - 1) - 1, -2 ** (bits - 1))

        return s

    def angle_from_points(self, x1, x2):

        return np.angle((x1[0, 0] - x2[0, 0]) + 1j * (x1[1, 0] - x2[1, 0]))

    def normalize_pwr(self, sig1, sig2):
        '''
        normalize sig1 to have the same power as sig2
        '''

        # average power per sample
        p1 = np.mean(sig1 ** 2)
        p2 = np.mean(sig2 ** 2)

        # normalize
        return sig1.copy() * np.sqrt(p2 / p1)

    def lowpass(self, signal, Fs, fc=fc_hp, plot=False):
        '''
        Filter out the really low frequencies, default is below 50Hz
        '''

        # have some predefined parameters
        rp = 5  # minimum ripple in dB in pass-band
        rs = 60  # minimum attenuation in dB in stop-band
        n = 4  # order of the filter
        type = 'butter'

        # normalized cut-off frequency
        wc = 2. * fc / Fs

        # design the filter
        from scipy.signal import iirfilter, lfilter, freqz
        b, a = iirfilter(n, Wn=wc, rp=rp, rs=rs, btype='lowpass', ftype=type)

        # plot frequency response of filter if requested
        if (plot):
            import matplotlib.pyplot as plt
            w, h = freqz(b, a)

            plt.figure()
            plt.title('Digital filter frequency response')
            plt.plot(w, 20 * np.log10(np.abs(h)))
            plt.title('Digital filter frequency response')
            plt.ylabel('Amplitude Response [dB]')
            plt.xlabel('Frequency (rad/sample)')
            plt.grid()

        # apply the filter
        signal = lfilter(b, a, signal.copy())

        return signal

    def highpass(self, signal, Fs, fc=fc_hp, plot=False):
        '''
        Filter out the really low frequencies, default is below 50Hz
        '''

        # have some predefined parameters
        rp = 5  # minimum ripple in dB in pass-band
        rs = 60  # minimum attenuation in dB in stop-band
        n = 4  # order of the filter
        type = 'butter'

        # normalized cut-off frequency
        wc = 2. * fc / Fs

        # design the filter
        from scipy.signal import iirfilter, lfilter, freqz
        b, a = iirfilter(n, Wn=wc, rp=rp, rs=rs, btype='highpass', ftype=type)

        # plot frequency response of filter if requested
        if (plot):
            import matplotlib.pyplot as plt
            w, h = freqz(b, a)

            plt.figure()
            plt.title('Digital filter frequency response')
            plt.plot(w, 20 * np.log10(np.abs(h)))
            plt.title('Digital filter frequency response')
            plt.ylabel('Amplitude Response [dB]')
            plt.xlabel('Frequency (rad/sample)')
            plt.grid()

        # apply the filter
        signal = lfilter(b, a, signal.copy())

        return signal

    def time_dB(self, signal, Fs, bits=16):
        '''
        Compute the signed dB amplitude of the oscillating signal
        normalized wrt the number of bits used for the signal
        '''

        import matplotlib.pyplot as plt

        # min dB (least significant bit in dB)
        lsb = -20 * np.log10(2.) * (bits - 1)

        # magnitude in dB (clipped)
        pos = self.clip(signal, 2. ** (bits - 1) - 1, 1.) / 2. ** (bits - 1)
        neg = -self.clip(signal, -1., -2. ** (bits - 1)) / 2. ** (bits - 1)

        mag_pos = np.zeros(signal.shape)
        Ip = np.where(pos > 0)
        mag_pos[Ip] = 20 * np.log10(pos[Ip]) + lsb + 1

        mag_neg = np.zeros(signal.shape)
        In = np.where(neg > 0)
        mag_neg[In] = 20 * np.log10(neg[In]) + lsb + 1

        plt.plot(np.arange(len(signal)) / float(Fs), mag_pos - mag_neg)
        plt.xlabel('Time [s]')
        plt.ylabel('Amplitude [dB]')
        plt.axis('tight')
        plt.ylim(lsb - 1, -lsb + 1)

        # draw ticks corresponding to decibels
        div = 20
        n = int(-lsb / div) + 1
        yticks = np.zeros(2 * n)
        yticks[:n] = lsb - 1 + np.arange(0, n * div, div)
        yticks[n:] = -lsb + 1 - np.arange((n - 1) * div, -1, -div)
        yticklabels = np.zeros(2 * n)
        yticklabels = range(0, -n * div, -div) + range(-(n - 1) * div, 1, div)
        plt.setp(plt.gca(), 'yticks', yticks)
        plt.setp(plt.gca(), 'yticklabels', yticklabels)

        plt.setp(plt.getp(plt.gca(), 'ygridlines'), 'ls', '--')

